
%{


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "jas.h"

int flag;

#define YYSTYPE unsigned
#define YYDEBUG 1 /* we use yyname[] */

%}

%token	ySTRING ySYMBOL yVALUE
%token	yDOT yORG yTEXT yDATA
%token	yRS yBYTE yWORD yASCII yASCIZ
%token	yHI yLO ySIZE yASSERT
%token	yALIGN yFILL yCKSUM yASG

%right	'='
%left	'+' '-'
%left	'*' '/' '%'
%left	yEQU yNEQ yLTE yGTE yOR yAND '>' '<'
%left	ySHL ySHR
%left	'|' '&' '^'
%left	yUPLUS yUMINUS '~'

%start program

%%

program		:
			| program labels line ';' {
				setsize();
			}
			;

line		:
			| directive
			| instruction
			;

labels		:
			| labels ySYMBOL ':' {
				define($2, dot, 'l');
			}
			;

directive	: yORG expr { (intext ? textdot : datadot) = dot = $2; }

			| yTEXT { intext = 1; dot = textdot; }
			| yDATA { intext = 0; dot = datadot; }
			
			| yCKSUM { cksum(); }

			| yASSERT '(' expr ')' {
				if (! ($$ = $3))
					yyerror("Assert");
			}
			| ySYMBOL yASG expr {
				define($1, $3, 'a');
			}
			| ySYMBOL '=' expr {
				define($1, $3, 'a');
			}
			| yRS {
				emit(NULL, 1);
			}
			| yRS expr {
				emit(NULL, $2);
			}
			| yFILL expr ',' expr {
				fill($4, $2);
			}
			| yALIGN expr {
				align(0, $2);
			}
			| yALIGN expr ',' expr {
				align($4, $2);
			}
			| yBYTE  { flag = 'b'; } exprlist
			| yWORD  { flag = 'w'; } exprlist
			| yASCII { flag = 'i'; } stringlist
			| yASCIZ { flag = 'z'; } stringlist
			;

stringlist	: ySTRING {
				emit($1, yyleng + (flag == 'z'));
			}
			| stringlist ',' ySTRING {
				emit($3, yyleng + (flag == 'z'));
			}
			;
exprlist	: expr {
				(flag == 'b' ? emitb : emitw)($1);
			}
			| exprlist ',' expr {
				(flag == 'b' ? emitb : emitw)($3);
			}
			;

expr		: yDOT {
				$$ = dot;
			}
			| ySYMBOL {
				$$ = symval($1);
			}
			| yVALUE {
				$$ = $1;
			}
			| yHI '(' expr ')' {
				$$ = $3 >> 8;
			}
			| yLO '(' expr ')' {
				$$ = $3 & 0xFF;
			}
			| ySIZE '(' ySYMBOL ')' {
				$$ = symsize($3);
			//}
			//| yCKSUM '(' expr ',' expr ')' {
			//	$$ = cksum($3, $5);

			}
			| '(' expr ')' {
				$$ = $2;
			}
			| '+' expr %prec yUPLUS  { $$ = + $2; }
			| '-' expr %prec yUMINUS { $$ = - $2; }
			| '~' expr               { $$ = ~ $2; }

			| expr '|' expr { $$ = $1 | $3; }
			| expr '&' expr { $$ = $1 & $3; }
			| expr '^' expr { $$ = $1 ^ $3; }
			| expr '*' expr { $$ = $1 * $3; }
			| expr '/' expr { $$ = $1 / $3; }
			| expr '%' expr { $$ = $1 % $3; }
			| expr '+' expr { $$ = $1 + $3; }
			| expr '-' expr { $$ = $1 - $3; }

			| expr yEQU expr { $$ = $1 == $3; }
			| expr yNEQ expr { $$ = $1 != $3; }
			| expr '<'  expr { $$ = $1 <  $3; }
			| expr '>'  expr { $$ = $1 >  $3; }
			| expr yLTE expr { $$ = $1 <= $3; }
			| expr yGTE expr { $$ = $1 >= $3; }
			| expr yOR  expr { $$ = $1 || $3; }
			| expr yAND expr { $$ = $1 && $3; }

			| expr ySHL expr { $$ = $1 << $3; }
			| expr ySHR expr { $$ = $1 >> $3; }

			;

wabs		: expr {
				$$ = $1;
			}
			;

brel		: expr {
				/*
				 *  Wierd (but consistent) way...
				 */
				$$ = $1 & 0xFF;
				if (pass == 1 && (((dot + 1) & 0xFF00) | $$) != $1)
					yyerror("relative out of range");
				if (pass == 1 && (dot & 0xFF) == 0xFF)
					yyerror("boundary branch");
#if 0
					fprintf(stderr, "Warning: boundary branch\n");
#endif
			}
			;

Dn			: expr {
				$$ = $1;
				if (pass == 1 && ($$ < 1 || $$ > 7))
					yyerror("io out of range");
			}
			;

Rn			: R0 { $$ = 0x0; }
			| R1 { $$ = 0x1; }
			| R2 { $$ = 0x2; }
			| R3 { $$ = 0x3; }
			| R4 { $$ = 0x4; }
			| R5 { $$ = 0x5; }
			| R6 { $$ = 0x6; }
			| R7 { $$ = 0x7; }
			| R8 { $$ = 0x8; }
			| R9 { $$ = 0x9; }
			| RA { $$ = 0xA; }
			| RB { $$ = 0xB; }
			| RC { $$ = 0xC; }
			| RD { $$ = 0xD; }
			| RE { $$ = 0xE; }
			| RF { $$ = 0xF; }
			| expr {
				$$ = $1;
				if (pass == 1 && ($$ < 0 || $$ > 0xF))
					yyerror("register out of range");
			}
			;

immb		: expr {
				$$ = $1;
				/*
				 * Ugly hack to handle ~0xAA
				 */
				if (($$ & ~0xFF) == ~0xFF)
					$$ &= 0xFF;
				if (pass == 1 && ($$ < 0 || $$ > 0xFF))
					yyerror("immediate out of range");
			}
			;

instruction	: ADC			{ op (0x74); }
			| ADCI	immb	{ opb(0x7C , $2); }
			| ADD			{ op (0xF4); }
			| ADI	immb	{ opb(0xFC , $2); }
			| AND			{ op (0xF2); }
			| ANI	immb	{ opb(0xFA , $2); }
			| B1	brel	{ opb(0x34 , $2); }
			| B2	brel	{ opb(0x35 , $2); }
			| B3	brel	{ opb(0x36 , $2); }
			| B4	brel	{ opb(0x37 , $2); }
			| BDF	brel	{ opb(0x33 , $2); }
			| BN1	brel	{ opb(0x3C , $2); }
			| BN2	brel	{ opb(0x3D , $2); }
			| BN3	brel	{ opb(0x3E , $2); }
			| BN4	brel	{ opb(0x3F , $2); }
			| BNF	brel	{ opb(0x3B , $2); }
			| BNQ	brel	{ opb(0x39 , $2); }
			| BNZ	brel	{ opb(0x3A , $2); }
			| BQ	brel	{ opb(0x31 , $2); }
			| BR	brel	{ opb(0x30 , $2); }
			| BZ	brel	{ opb(0x32 , $2); }
			| DEC	Rn		{ op (0x20 | $2); }
			| DIS			{ op (0x71); }
			| GHI	Rn		{ op (0x90 | $2); }
			| GLO	Rn		{ op (0x80 | $2); }
			| IDL			{ op (0x00); }
			| INC	Rn		{ op (0x10 | $2); }
			| INP	Dn		{ op (0x68 | $2); }
			| IRX			{ op (0x60); }
			| LBDF	wabs	{ opw(0xC3 , $2); }
			| LBNF	wabs	{ opw(0xCB , $2); }
			| LBNQ	wabs	{ opw(0xC9 , $2); }
			| LBNZ	wabs	{ opw(0xCA , $2); }
			| LBQ	wabs	{ opw(0xC1 , $2); }
			| LBR	wabs	{ opw(0xC0 , $2); }
			| LBZ	wabs	{ opw(0xC2 , $2); }
			| LDA	Rn		{ op (0x40 | $2); }
			| LDI	immb	{ opb(0xF8 , $2); }
			| LDN	Rn		{ op (0x00 | $2); }
			| LDX			{ op (0xF0); }
			| LDXA			{ op (0x72); }
			| LSDF			{ op (0xCF); }
			| LSIE			{ op (0xCC); }
			| LSKP			{ op (0xC8); }
			| LSNF			{ op (0xC7); }
			| LSNQ			{ op (0xC5); }
			| LSNZ			{ op (0xC6); }
			| LSQ			{ op (0xCD); }
			| LSZ			{ op (0xCE); }
			| MARK			{ op (0x79); }
			| NOP			{ op (0xC4); }
			| OR			{ op (0xF1); }
			| ORI	immb	{ opb(0xF9 , $2); }
			| OUT	Dn		{ op (0x60 | $2); }
			| PHI	Rn		{ op (0xB0 | $2); }
			| PLO	Rn		{ op (0xA0 | $2); }
			| REQ			{ op (0x7A); }
			| RET			{ op (0x70); }
			| SAV			{ op (0x78); }
			| SD			{ op (0xF5); }
			| SDB			{ op (0x75); }
			| SDBI	immb	{ opb(0x7D , $2); }
			| SDI	immb	{ opb(0xFD , $2); }
			| SEP	Rn		{ op (0xD0 | $2); }
			| SEQ			{ op (0x7B); }
			| SEX	Rn		{ op (0xE0 | $2); }
			| SHL			{ op (0xFE); }
			| SHLC			{ op (0x7E); }
			| SHR			{ op (0xF6); }
			| SHRC			{ op (0x76); }
			| SKP			{ op (0x38); }
			| SM			{ op (0xF7); }
			| SMB			{ op (0x77); }
			| SMBI	immb	{ opb(0x7F , $2); }
			| SMI	immb	{ opb(0xFF , $2); }
			| STR	Rn		{ op (0x50 | $2); }
			| STXD			{ op (0x73); }
			| XOR			{ op (0xF3); }
			| XRI	immb	{ opb(0xFB , $2); }

			| STPC             { op2(0x68, 0x00); }
			| DTC              { op2(0x68, 0x01); }
			| SPM2             { op2(0x68, 0x02); }
			| SCM2             { op2(0x68, 0x03); }
			| SPM1             { op2(0x68, 0x04); }
			| SCM1             { op2(0x68, 0x05); }
			| LDC              { op2(0x68, 0x06); }
			| STM              { op2(0x68, 0x07); }
			| GEC              { op2(0x68, 0x08); }
			| ETQ              { op2(0x68, 0x09); }
			| XIE              { op2(0x68, 0x0A); }
			| XID              { op2(0x68, 0x0B); }
			| CIE              { op2(0x68, 0x0C); }
			| CID              { op2(0x68, 0x0D); }

			| DBNZ Rn ',' wabs { op4(0x68, 0x20 | $2, $4 >> 8, $4); }

			| BCI brel         { op3(0x68, 0x3E, $2); }
			| BXI brel         { op3(0x68, 0x3F, $2); }

			| RLXA Rn          { op2(0x68, 0x60 | $2); }

			| DADC             { op2(0x68, 0x74); }
			| DSAV             { op2(0x68, 0x76); }
			| DSMB             { op2(0x68, 0x77); }
			| DACI immb        { op3(0x68, 0x7C, $2); }
			| DSBI immb        { op3(0x68, 0x7F, $2); }

			| SCAL Rn ',' wabs { op4(0x68, 0x80 | $2, $4 >> 8, $4); }
			| SRET Rn          { op2(0x68, 0x90 | $2); }

			| RSXD Rn          { op2(0x68, 0xA0 | $2); }
			| RNX  Rn          { op2(0x68, 0xB0 | $2); }
			| RLDI Rn ',' wabs { op4(0x68, 0xC0 | $2, $4 >> 8, $4); }

			| DADD             { op2(0x68, 0xF4); }
			| DSM              { op2(0x68, 0xF7); }
			| DADI immb        { op3(0x68, 0xFC, $2); }
			| DSMI immb        { op3(0x68, 0xFF, $2); }

			;

%%

int
getkeyword(name)
	char *name;
{
	int i;

	for (i = YYERRCODE + 1; i <= YYMAXTOKEN; ++i)
		if (isupper(*yyname[i]) && !strcasecmp(yyname[i], name))
			return (i);


#define MAP(x,y) if (!strcmp(name, x)) return (y);

	MAP(".",      yDOT)

	MAP(".rs",    yRS)
	MAP(".org",   yORG)
	MAP(".text",  yTEXT)
	MAP(".data",  yDATA)
	MAP(".byte",  yBYTE)
	MAP(".word",  yWORD)
	MAP(".ascii", yASCII)
	MAP(".asciz", yASCIZ)
	MAP(".align", yALIGN)
	MAP(".fill",  yFILL)
	MAP(".cksum", yCKSUM)
	MAP("equ",    yASG)

	MAP("HI",     yHI)
	MAP("LO",     yLO)
	MAP("SIZE",   ySIZE)
	MAP("ASSERT", yASSERT)

	return (0);
}


static char *vers2 = "@(#)cdp1806.y 970514";

char *cpu_family = "CDP1806";
int MSB_first = 0;


